#!/usr/bin/env python

import sys
from optparse import OptionParser
import lex
import yacc


#
# Constants
# For each new node type, add a new constant here.
#

NodeType_none =                      0
NodeType_elementpattern =            1
NodeType_textpattern =               2
NodeType_pattern =                   3
NodeType_grammar =                   4
NodeType_sequence =                  5
NodeType_ident =                     6
NodeType_zeroormore =                7
NodeType_oneormore =                 8
NodeType_zeroorone =                 9
NodeType_commasequence =            10
NodeType_ampersandsequence =        11
NodeType_barsequence =              12
NodeType_attributepattern =         13
NodeType_identpattern =             14
NodeType_toplevel =                 15
NodeType_grammarcontentsequence =   16
NodeType_grammarcontent =           17
NodeType_start =                    18
NodeType_define =                   19
NodeType_assignmethod =             20
NodeType_parenpattern =             21
NodeType_emptypattern =             22
NodeType_listpattern =              23
NodeType_parentpattern =            24
NodeType_mixedpattern =             25
NodeType_notallowed =               26
NodeType_literal =                  27
NodeType_external =                 28
NodeType_inherit =                  29
NodeType_exceptpattern =            30
NodeType_datatype =                 31
NodeType_datatypevalue =            32
NodeType_datatypename =             33
NodeType_paramlist =                34
NodeType_param =                    35

#
# For each new node type, add another entry to this dictionary.
#
NodeType2NameDict = {
    NodeType_none: 'None',
    NodeType_elementpattern: 'ElementPattern',
    NodeType_textpattern: 'TextPattern',
    NodeType_emptypattern: 'EmptyPattern',
    NodeType_pattern: 'Pattern',
    NodeType_grammar: 'Grammar',
    NodeType_sequence: 'Sequence',
    NodeType_ident: 'Ident',
    NodeType_zeroormore: 'ZeroOrMore',
    NodeType_oneormore: 'OneOrMore',
    NodeType_zeroorone: 'ZeroOrOne',
    NodeType_commasequence: 'CommaSequence',
    NodeType_ampersandsequence: 'AmpersandSequence',
    NodeType_barsequence: 'BarSequence',
    NodeType_attributepattern: 'AttributePattern',
    NodeType_identpattern: 'IdentPattern',
    NodeType_toplevel: 'TopLevel',
    NodeType_grammarcontentsequence: 'GrammarContentSequence',
    NodeType_grammarcontent: 'GrammarContent',
    NodeType_start: 'Start',
    NodeType_define: 'Define',
    NodeType_assignmethod: 'AssignMethod',
    NodeType_parenpattern: 'ParenPattern',
    NodeType_listpattern: 'ListPattern',
    NodeType_parentpattern: 'ParentPattern',
    NodeType_mixedpattern: 'MixedPattern',
    NodeType_notallowed: 'NotAllowedPattern',
    NodeType_literal: 'Literal',
    NodeType_external: 'External',
    NodeType_inherit: 'Inherit',
    NodeType_exceptpattern: 'ExceptPattern',
    NodeType_datatype: 'DataType',
    NodeType_datatypevalue: 'DataTypeValue',
    NodeType_datatypename: 'DataTypeName',
    NodeType_paramlist: 'ParamList',
    NodeType_param: 'Param',
}


#
# Representation of a node in the AST (abstract syntax tree).
#
class ASTNode:
    def __init__(self, nodeType, *children):
        self.nodeType = nodeType
        self.value = None
        self.children = list(children)

    def getChildren(self): return self.children
    def setChildren(self, children): self.children = children
    def getType(self): return self.nodeType
    def setType(self, nodeType): self.nodeType = nodeType
    def getValue(self): return self.value
    def setValue(self, value): self.value = value

    def addChild(self, newchild):
        self.children.append(newchild)
    def insertChild(self, index, newchild):
        self.children.insert(index, newchild)

    def __repr__(self):
        content = []
        s1 = '<Node type: %s  children: [' % (NodeType2NameDict[self.nodeType],)
        content.append(s1)
        firstTime = True
        for child in self.children:
            if firstTime:
                s1 = '\n        %s' % NodeType2NameDict[child.getType()]
                firstTime = False
            else:
                s1 = ',\n        %s' % NodeType2NameDict[child.getType()]
            content.append(s1)
        s1 = ']>'
        content.append(s1)
        s1 = ''.join(content)
        return s1

    def show(self, outstream, level):
        self.show_level(outstream, level)
        if self.children:
            childCount = len(self.children)
        else:
            childCount = 0
        if self.value is None:
            value = 'None'
        else:
            value = '"%s"' % self.value
        outstream.write('Type: %s  Value: %s  Child #: %d\n' % (
            NodeType2NameDict[self.nodeType],
            value,
            childCount))
        for child in self.children:
            child.show(outstream, level + 1)

    def show_level(self, outstream, level):
        for idx in range(level):
            outstream.write('  ')


#
# Exception classes
#
class LexerError(Exception):
    def __init__(self, msg, lineno, columnno):
        self.msg = msg
        self.lineno = lineno
        self.columnno = columnno
    def show(self):
        sys.stderr.write('Lexer error (%d, %d) %s\n' % \
            (self.lineno, self.columnno, self.msg))

class ParserError(Exception):
    def __init__(self, msg, lineno, columnno):
        self.msg = msg
        self.lineno = lineno
        self.columnno = columnno
    def show(self):
        sys.stderr.write('Parser error (%d, %d) %s\n' % \
            (self.lineno, self.columnno, self.msg))


#
# Lexer specification
#
# For each new token type, add an entry here.
#
tokens = (
    'ELEMENT',
    'ATTRIBUTE',
    'TEXT',
    'LBRAC',
    'RBRAC',
    'LPAREN',
    'RPAREN',
    'COMMA',
    'AMPERSAND',
    'BAR',
    'IDENT',
    'STAR',
    'QUESTION',
    'PLUS',
    'MINUS',
    'EQUAL',
    'BAREQUAL',
    'AMPERSANDEQUAL',
    'START',
    'GRAMMAR',
    'EMPTY',
    'LIST',
    'PARENT',
    'MIXED',
    'NOTALLOWED',
    'LITERAL',
    'EXTERNAL',
    'INHERIT',
    'STRING',
    'TOKEN',
    'COLON',
    )

#
# Punctuation (etc ?).
# for each new punctuation, add a line here.
#
t_LBRAC   = r'\{'
t_RBRAC   = r'\}'
t_LPAREN   = r'\('
t_RPAREN   = r'\)'
t_COMMA  = r'\,'
t_AMPERSAND  = r'\&'
t_BAR  = r'\|'
t_STAR = r'\*'
t_QUESTION = r'\?'
t_PLUS = r'\+'
t_MINUS = r'\-'
t_EQUAL = r'='
t_BAREQUAL = r'\|='
t_AMPERSANDEQUAL = r'\&='
t_COLON = r'\:'
#t_IDENT    = r'[a-zA-Z_][a-zA-Z0-9_]*'
t_LITERAL = r'"[^\"\n]*"'

#
# Reserved keywords
# For each new keyword, add an entry to this dictionary.
#
reserved = {
    'element' : 'ELEMENT',
    'text' : 'TEXT',
    'attribute': 'ATTRIBUTE',
    'start': 'START',
    'grammar': 'GRAMMAR',
    'empty': 'EMPTY',
    'list': 'LIST',
    'parent': 'PARENT',
    'mixed': 'MIXED',
    'notAllowed': 'NOTALLOWED',
    'external': 'EXTERNAL',
    'inherit': 'INHERIT',
    'string': 'STRING',
    'token': 'TOKEN',
}

def t_IDENT(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'IDENT')
    return t

t_ignore = " \t"

# Comments - Skip them, but cound the newline.
def t_comment(t):
    r'\#.*\n'
    t.lineno += 1

# Newlines - Skip them, but count them.
def t_newline(t):
    r'\n+'
    t.lineno += t.value.count("\n")

def t_error(t):
    print "Illegal character '%s'" % t.value[0]
    t.skip(1)


#
# Define the grammar.
#

def p_toplevel(t):
    'toplevel : pattern'
    node = ASTNode(NodeType_toplevel, t[1])
    t[0] = node

def p_pattern_1(t):
    'pattern : element'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_2(t):
    'pattern : text_pattern'
    t[0] = t[1]

def p_pattern_empty(t):
    'pattern : empty_pattern'
    t[0] = t[1]

def p_pattern_3(t):
    'pattern : pattern STAR'
    node = ASTNode(NodeType_zeroormore, t[1])
    t[0] = node

def p_pattern_4(t):
    'pattern : pattern PLUS'
    node = ASTNode(NodeType_oneormore, t[1])
    t[0] = node

def p_pattern_5(t):
    'pattern : pattern QUESTION'
    node = ASTNode(NodeType_zeroorone, t[1])
    t[0] = node

def p_pattern_6(t):
    'pattern : attribute'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_ident(t):
    'pattern : identpattern'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_grammar(t):
    'pattern : grammar'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_paren(t):
    'pattern : paren_pattern'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_list(t):
    'pattern : list_pattern'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_parent(t):
    'pattern : parent_pattern'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_mixed(t):
    'pattern : mixed_pattern'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_notallowed(t):
    'pattern : NOTALLOWED'
    node = ASTNode(NodeType_pattern)
    node1 = ASTNode(NodeType_notallowed)
    node.addChild(node1)
    t[0] = node

def p_pattern_external(t):
    'pattern : external'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_pattern_data_type(t):
    'pattern : data_type'
    node = ASTNode(NodeType_pattern, t[1])
    t[0] = node

def p_grammar(t):
    'grammar : GRAMMAR LBRAC grammar_content_seq RBRAC'
    node = ASTNode(NodeType_grammar, t[3])
    t[0] = node

def p_grammar_content_seq_1(t):
    'grammar_content_seq : grammar_content'
    node = ASTNode(NodeType_grammarcontentsequence, t[1])
    t[0] = node

def p_grammar_content_seq_2(t):
    'grammar_content_seq : grammar_content_seq grammar_content'
    t[1].addChild(t[2])
    t[0] = t[1]

def p_grammar_content_1(t):
    'grammar_content : start'
    node = ASTNode(NodeType_grammarcontent, t[1])
    t[0] = node

def p_grammar_content_2(t):
    'grammar_content : define'
    node = ASTNode(NodeType_grammarcontent, t[1])
    t[0] = node

def p_start(t):
    'start : START assign_method pattern'
    node = ASTNode(NodeType_start, t[2], t[3])
    t[0] = node

def p_define(t):
    'define : ident assign_method pattern_seq'
    node = ASTNode(NodeType_define, t[1], t[2], t[3])
    t[0] = node

def p_assign_method(t):
    '''assign_method : EQUAL
                   | BAREQUAL
                   | AMPERSANDEQUAL'''
    node = ASTNode(NodeType_assignmethod)
    node.setValue(t[1])
    t[0] = node

def p_paren_pattern(t):
    'paren_pattern : LPAREN pattern_seq RPAREN'
    node = ASTNode(NodeType_parenpattern, t[2])
    t[0] = node

def p_parent_pattern(t):
    'parent_pattern : PARENT ident'
    node = ASTNode(NodeType_parentpattern, t[2])
    t[0] = node

def p_mixed_pattern(t):
    'mixed_pattern : MIXED LBRAC pattern_seq RBRAC'
    node = ASTNode(NodeType_mixedpattern, t[3])
    t[0] = node

def p_list_pattern(t):
    'list_pattern : LIST LBRAC pattern_seq RBRAC'
    node = ASTNode(NodeType_listpattern, t[3])
    t[0] = node

def p_external_1(t):
    'external : EXTERNAL anyURIliteral'
    node = ASTNode(NodeType_external, t[2])
    t[0] = node

def p_external_2(t):
    'external : EXTERNAL anyURIliteral inherit'
    node = ASTNode(NodeType_external, t[2], t[3])
    t[0] = node

def p_anyURIliteral(t):
    'anyURIliteral : literal'
    t[0] = t[1]

def p_literal(t):
    'literal : LITERAL'
    node = ASTNode(NodeType_literal)
    node.setValue(t[1])
    t[0] = node

def p_inherit(t):
    'inherit : INHERIT EQUAL ident'
    node = ASTNode(NodeType_inherit, t[3])
    t[0] = node

def p_text_pattern(t):
    'text_pattern : TEXT'
    node = ASTNode(NodeType_textpattern)
    t[0] = node

def p_empty_pattern(t):
    'empty_pattern : EMPTY'
    node = ASTNode(NodeType_emptypattern)
    t[0] = node

def p_element(t):
    'element : ELEMENT ident LBRAC pattern_seq RBRAC'
    node = ASTNode(NodeType_elementpattern, t[2], t[4])
    t[0] = node

def p_attribute(t):
    'attribute : ATTRIBUTE ident LBRAC pattern_seq RBRAC'
    node = ASTNode(NodeType_attributepattern, t[2], t[4])
    t[0] = node

def p_identpattern(t):
    'identpattern : ident'
    node = ASTNode(NodeType_identpattern)
    node.addChild(t[1])
    t[0] = node

def p_pattern_seq(t):
    'pattern_seq : pattern pattern_seq_tail'
    node = t[2]
    node.insertChild(0, t[1])
    t[0] = node

def p_pattern_seq_tail_0(t):
    'pattern_seq_tail : empty'
    node = ASTNode(NodeType_commasequence)
    t[0] = node

def p_pattern_seq_tail_1(t):
    '''pattern_seq_tail : pattern_comma_seq
                        | pattern_ampersand_seq
                        | pattern_bar_seq'''
    t[0] = t[1]

def p_pattern_comma_seq_1(t):
    'pattern_comma_seq : COMMA pattern'
    node = ASTNode(NodeType_commasequence, t[2])
    t[0] = node

def p_pattern_comma_seq_2(t):
    'pattern_comma_seq : pattern_comma_seq COMMA pattern'
    t[1].addChild(t[3])
    t[0] = t[1]

def p_pattern_ampersand_seq_1(t):
    'pattern_ampersand_seq : AMPERSAND pattern'
    node = ASTNode(NodeType_ampersandsequence, t[2])
    t[0] = node

def p_pattern_ampersand_seq_2(t):
    'pattern_ampersand_seq : pattern_ampersand_seq AMPERSAND pattern'
    t[1].addChild(t[3])
    t[0] = t[1]

def p_pattern_bar_seq_1(t):
    'pattern_bar_seq : BAR pattern'
    node = ASTNode(NodeType_barsequence, t[2])
    t[0] = node

def p_pattern_bar_seq_2(t):
    'pattern_bar_seq : pattern_bar_seq BAR pattern'
    t[1].addChild(t[3])
    t[0] = t[1]

def p_ident(t):
    'ident : IDENT'
    node = ASTNode(NodeType_ident)
    node.setValue(t[1])
    t[0] = node

def p_data_type_1(t):
    'data_type : data_type_value'
    node = ASTNode(NodeType_datatype, t[1])
    t[0] = node

def p_data_type_3(t):
    'data_type : data_type_name data_type_value'
    node = ASTNode(NodeType_datatype, t[1], t[2])
    t[0] = node

def p_data_type_4(t):
    'data_type : data_type_name brac_param_list except_pattern'
    node = ASTNode(NodeType_datatype, t[1], t[2], t[3])
    t[0] = node

def p_data_type_value(t):
    'data_type_value : literal'
    node = ASTNode(NodeType_datatypevalue, t[1])
    t[0] = node

def p_data_type_name(t):
    '''data_type_name : CName
                      | STRING
                      | TOKEN'''
    node = ASTNode(NodeType_datatypename)
    node.setValue(t[1])
    t[0] = node

def p_cname(t):
    'CName : ident COLON ident'
    # Throw away the ident nodes and return a string.
    value = '%s:%s' % (t[1].getValue(), t[3].getValue())
    t[0] = value

def p_brac_param_list_1(t):
    'brac_param_list : empty'
    node = ASTNode(NodeType_paramlist)
    t[0] = node

def p_brac_param_list_2(t):
    'brac_param_list : LBRAC param_list RBRAC'
    t[0] = t[2]

def p_param_list_1(t):
    'param_list : empty'
    node = ASTNode(NodeType_paramlist)
    t[0] = node

def p_param_list_2(t):
    'param_list : param'
    node = ASTNode(NodeType_paramlist, t[1])
    t[0] = node

def p_param_list_3(t):
    'param_list : param_list param'
    t[1].addChild(t[2])
    t[0] = t[1]

def p_param(t):
    'param : ident EQUAL literal'
    node = ASTNode(NodeType_param, t[1], t[3])
    t[0] = node

def p_except_pattern_1(t):
    'except_pattern : empty'
    node = ASTNode(NodeType_exceptpattern)
    t[0] = node

def p_except_pattern_2(t):
    'except_pattern : MINUS pattern'
    node = ASTNode(NodeType_exceptpattern, t[2])
    t[0] = node

def p_empty(t):
    'empty :'
    pass

def p_error(t):
    print "Syntax error at '%s:%d'" % (t.value, t.lineno)
    #print dir(t)


#
# Run the test.
# Parse a file and display the tree.
#
def parse(infilename, debug=0):
    lex.lex(debug=debug)
    yacc.yacc(debug=debug)
    infile = file(infilename, 'r')
    content = infile.read()
    infile.close()
    try:
        # Do the parse
        result = yacc.parse(content)
        # Display the AST
        if result:
            return result
        else:
            sys.stderr.write('Error attempting to parse file: %s\n' % infilename)
    except LexerError, exp:
        exp.show()
    except ParserError, exp:
        exp.show()


#
# Run the test.
# Parse a file and display the tree.
#
def parse_and_show(infilename, debug=0):
    lex.lex(debug=debug)
    yacc.yacc(debug=debug)
    infile = file(infilename, 'r')
    content = infile.read()
    infile.close()
    try:
        # Do the parse
        #raw_input('Start parse ')
        result = yacc.parse(content)
        #raw_input('Finished parse ')
        # Display the AST
        if result:
            result.show(sys.stdout, 0)
    except LexerError, exp:
        exp.show()
    except ParserError, exp:
        exp.show()


USAGE_TEXT = """
    %prog [options] <infile.rng>
example:
    %prog infile.rng"""

def usage(parser):
    parser.print_help()
    sys.exit(-1)

def main():
    parser = OptionParser(usage=USAGE_TEXT, version='1.0a',
        description='Parse RELAX NG compact syntax.')
##     parser.add_option("-f", "--file", type="string", dest="filename",
##         help="read data from FILENAME")
    parser.add_option("-d", "--debug", help="run lexer and parser in debug mode",
        action="store_true", dest="debug")
##     parser.add_option("-q", "--quiet",
##         action="store_false", dest="verbose")
    (options, args) = parser.parse_args()
    if len(args) != 1:
        usage(parser)
    infilename = args[0]
    debug = 0
    if options.debug:
        debug = 1
    parse_and_show(infilename, debug)


if __name__ == "__main__":
    main()
    #import pdb
    #pdb.run('main()')

